
########################################### FUNCTIONS TO FIT PHYLOGENETIC MODEL ##################################################


solve.for.lik <- function(Sigma, VL, X){ # solve for likelihood given var-covar matrix Sigma, data VL, design matrix X (for fixed effects)
  n <- length(VL)
  stopifnot(nrow(Sigma) == n)
  stopifnot(nrow(X) == n)
  lik <- - 10^9
  beta <- NULL
  
  tryCatch( # catch error
{
  
  if(all(Sigma[lower.tri(Sigma)] == 0, Sigma[upper.tri(Sigma)] == 0)){ # if matrix is diagonal:
    inverse.Sigma <- diag(n)
    diag(inverse.Sigma) <- 1/diag(Sigma)
    log.det.Sigma <- sum(log(diag(Sigma)))
  } else { # non-diagonal matrix use slower functions
    inverse.Sigma <- chol2inv(chol(Sigma)) # inverse from cholesky decomposition is faster
    log.det.Sigma <- determinant(Sigma, logarithm=T)$modulus[1]
  }  
  beta <- solve(t(X) %*% inverse.Sigma %*% X)  %*% (t(X) %*% inverse.Sigma %*% VL) # gives you best intercept and slope
  least.square.mu <- X %*% beta # average at the tips according to least square regression
  lik <- -(1/2) * (n * log(2*pi) + log.det.Sigma + t(VL- X %*% beta) %*% inverse.Sigma %*% (VL- X %*% beta))
}
,
error = function(e) {
  print(e)
}
  )
return(list(beta = beta, loglik = lik))
}
get.ML.par <- function(par, var, VL, model, mm){
  
  tmp <- create.SigmaG.SigmaE(par, var, model, mm)
  Sigma <- tmp[[1]] + tmp[[2]]
  X <- tmp[[3]]
  tmp <- solve.for.lik(Sigma, VL, X)
  return(tmp)
}

create.SigmaG.SigmaE <- function(par, var, model, mm = NULL){ # create covariance matrices and design matrix
  stopifnot(model == "white" | model == "BM" | model == "OU" | model == "BM3" | model == "BM.VEt" | model == "VEt")
  
  ntip <- nrow(var[[1]])
  if(!is.null(mm) & all(mm[,1] == 1)) stop("constant must be removed from model matrix")
  if(!is.null(mm)) if(nrow(mm) != ntip) stop("model matrix does not have the right number of rows")
  # for each model create appropriate Sigma and design matrix X
  
  if(model == "white"){
    if(length(par) != 1) stop("white has 1 parameter")
    SigmaG <- matrix(0, ntip, ntip)
    SigmaE <- par[1] * diag(ntip)
    X <- matrix(rep(1, ntip), ncol = 1)
  }
  if(model == "BM"){
    if(length(par) != 2) stop("BM has 2 parameters")
    if(length(var) != 1) stop("BM has 1 variable, the var-covar matrix")
    VE <- par[1]
    sigsq <- par[2]
    vcv.mat <- var[[1]]
    SigmaG <- sigsq * vcv.mat
    SigmaE <- VE * diag(ntip)  # analytical formula for variance-covariance matrix for BM
    X <- matrix(rep(1, ntip), ncol = 1) # design matrix for BM model
  }
  if(model == "OU"){
    if(length(par) != 3) stop("OU has 3 parameters")
    if(length(var) != 2) stop("OU has 2 variables, the var-covar matrix and the patristic distance matrix")
    VE <- par[1]
    sigsq <- par[2]
    alpha <- par[3]
    vcv.mat <- var[[1]]
    pat.dist.mat <- var[[2]]
    root.to.tip <- diag(vcv.mat)
    SigmaG <- (sigsq / (2*alpha)) * (1 - exp(-2*alpha*vcv.mat)) * exp(-alpha * pat.dist.mat) # analytical formula for variance-covariance matrix
    SigmaE <- VE * diag(ntip)
    X <- cbind(exp(-alpha * root.to.tip), 1-exp(-alpha * root.to.tip))# the design matrix, zA and theta (in this order)
  }
  if(model == "BM3"){
    if(length(par) != 4) stop("BM3 has 4 parameters")
    if(length(var) != 1) stop("BM3 has 1 variable, the var-covar matrix")
    VE <- par[1]
    sigsq0 <- par[2]
    sigsq1 <- par[3]
    tlim <- par[4]
    vcv.mat <- var[[1]]
    
    is.smaller <- as.numeric(vcv.mat <= tlim)
    SigmaG <- sigsq0 * vcv.mat * is.smaller + (sigsq0 * tlim + sigsq1 * (vcv.mat - tlim)) * (1-is.smaller)
    SigmaE <- VE * diag(ntip)
    X <- matrix(rep(1, ntip), ncol = 1) # the design matrix for BM model
  }
  if(model == "BM.VEt"){
    if(length(par) != 3) stop("BM.VEt has 3 parameters")
    if(length(var) != 2) stop("BM.VEt has 2 variables, the var-covar matrix and the vector of dates")
    VE0 <- par[1]
    sigsq <- par[2]
    slope.VE <- par[3]
    vcv.mat <- var[[1]]
    dates <- var[[2]]    
    SigmaG <- sigsq * vcv.mat # analytical formula for variance-covariance matrix for BM
    SigmaE <- (VE0 + dates * slope.VE) * diag(ntip)
    X <- matrix(rep(1, ntip), ncol = 1) # the design matrix for BM model
  }
  if(model == "VEt"){
    if(length(par) != 2) stop("VEt has 2 parameters")
    if(length(var) != 1) stop("VEt has 1 variable, the vector of dates")
    VE0 <- par[1]
    slope.VE <- par[2]
    dates <- var[[1]]
    SigmaG <- matrix(0, ntip, ntip)
    SigmaE <- (VE0 + dates * slope.VE) * diag(ntip)
    X <- matrix(rep(1, ntip), ncol = 1) # the design matrix
  }
  # add other fixed effects
  if(!is.null(mm)){
    X <- cbind(X, mm) # additional fixed effects
  }
  return(list(SigmaG, SigmaE, X))
}

get.simulated.trait <- function(VL, pars, vars, model, mm, n.sim){ # get h2 for a given VL and parameter, variables, and model
  n.obs  <- length(VL)
  # get the Sigma and design matrix
  tmp <- create.SigmaG.SigmaE(par = pars, var = vars, model =  model, mm = mm)
  SigmaG <- tmp[[1]]
  SigmaE <- tmp[[2]]
  X <- tmp[[3]]
  mu <- X %*% get.ML.par(par = pars, var = vars, VL = VL, model = model, mm = mm)$beta # average at the tips according to least square regression
  
  if(all(SigmaG >= 0)){
    sim.g <- mvrnorm(n = n.sim, mu = mu, Sigma = SigmaG,
                     tol = 1e-6, empirical = FALSE, EISPACK = FALSE)
    sim.e <- mvrnorm(n = n.sim, mu = rep(0, n.obs), Sigma = SigmaE,
                     tol = 1e-6, empirical = FALSE, EISPACK = FALSE)
    sim.pheno <- sim.g + sim.e # 1000 trajectories * ntips
    return(sim.pheno)
  } else {
    return(NA)    
  }
}

##### NEW OPTIMISED LIKELIHOOD FUNCTIONS #####

lik.OU <- function(par, tree, VL, D, root.to.tip, full.output = F, mm = NULL){ # special fast OU function based on phylolm package, Lam Ho and Cecile Ane
  stopifnot(is.phylo(tree))
  stopifnot(length(par) == 3)
  
  n <- length(VL)
  stopifnot(length(root.to.tip) == n)
  stopifnot(length(D) == n)
  stopifnot(nrow(mm) == n) # number of rows of model matrix
  if(!is.null(mm) & all(mm[,1] == 1)) stop("constant must be removed from model matrix")
  
  VE <- par[1]
  sigsq <- par[2]
  alpha <- par[3]
  #print(par)
  
  tr <- transf.branch.lengths(phy = tree, model = "OUfixedRoot", parameters = list(alpha = alpha, sigma2_error = VE, sigma2 = sigsq),
                               check.pruningwise = F, check.ultrametric = F, D = D) # same with an experimental error and with sigma2

  X <- cbind(exp(-alpha * root.to.tip), 1-exp(-alpha * root.to.tip)) # for OU, zA and theta (in this order)
  
  if(!is.null(mm)){ # add other fixed effects
    X <- cbind(X, mm) # additional fixed effects
  }
  rownames(X) <- names(VL)
  
  result <- three.point.compute(tr$tree, P = X, Q = VL, tr$diagWeight) # with sigma2_error and sigma2
  beta <- solve(result$PP) %*% result$QP
  det <- result$logd
  ho.lik <- (1/2) * (n * log(2*pi) + n * (log(sigsq) - log(2 *alpha)) + det + 2 * alpha / sigsq * (result$QQ - result$QP %*% beta))
  if(full.output){
    return(list(lik = ho.lik, beta = beta))
  } else {
    return(ho.lik)
  }
}
lik.BM <- function(par, tree, VL, full.output = F, mm = NULL){ # special fast BM function based on phylolm package, Lam Ho and Cecile Ane
  stopifnot(is.phylo(tree))
  stopifnot(length(par) == 2)
  n <- length(VL)
  stopifnot(nrow(mm) == n) # number of rows of model matrix
  if(!is.null(mm) & all(mm[,1] == 1)) stop("constant must be removed from model matrix")
  
  VE <- par[1]
  sigsq <- par[2]

  #print(par)
  
  tr <- transf.branch.lengths(phy = tree, model = "BM", parameters = list(sigma2_error = VE, sigma2 = sigsq),
                              check.pruningwise = F, check.ultrametric = NULL) # same with an experimental error and with sigma2. For BM check.ultrametric not needed
  
  
  X <- matrix(rep(1, n), ncol = 1) # the design matrix for BM model
  if(!is.null(mm)){ # add other fixed effects
    X <- cbind(X, mm) # additional fixed effects
  }
  rownames(X) <- names(VL)
  result <- three.point.compute(tr$tree, P = X, Q = VL, tr$diagWeight)

  beta <- solve(result$PP) %*% result$QP # ancestral trait
  det <- result$logd
  
  ho.lik <- (1/2) * (n * log(2*pi) + n * log(sigsq) + det + 1 / sigsq * (result$QQ - result$QP %*% beta))
  if(full.output){
    return(list(lik = ho.lik, beta = beta))
  } else {
    return(ho.lik)
  }
}
lik.all <- function(par, var, VL, model){
  
  # do a number of checks on parameters
  if(par[1] < 0) return(10^9) # VE has to be positive
  if((model == "BM" | model == "OU" | model == "BM3" | model == "BM.VEt") & par[2] < 0) return(10^9) # sigma2 has to be positive
  if(model == "BM3" & par[3] < 0) return(10^9)
  if(model == "OU" & par[3] < 0) return(10^9)
  
  tmp <- create.SigmaG.SigmaE(par, var, model)
  Sigma <- tmp[[1]] + tmp[[2]]
  X <- tmp[[3]]
  
  if(any(Sigma < 0)) return(10^9)
  lik <- solve.for.lik(Sigma, VL, X)$loglik
  return(-lik)
}
clean.OU <- function(subtree, n){
  treeOU <- reorder(subtree, "pruningwise") # need to reorder the tree
  D <- numeric(n)
  des <- treeOU$edge[, 2]
  externalEdge <- (des <= n)
  root.to.tip <- pruningwise.distFromRoot(treeOU)[1:n]
  if (!is.ultrametric(treeOU)) {
    flag <- 1
    dis <- pruningwise.distFromRoot(treeOU) # all root to tip distances, and root to node distances
    names(dis) <- treeOU$tip.label
    D <- max(dis[1:n]) - dis[1:n]
    D <- D - mean(D)
    treeOU$edge.length[externalEdge] <- treeOU$edge.length[externalEdge] + D[des[externalEdge]]
  }
  return(list(treeOU = treeOU, root.to.tip = root.to.tip, D = D))
}
##### FUNCTIONS FOR BOOTSTRAP CONFIDENCE INTERVALS #####

do.bs <- function(VL, pars, vars, tree, D, root.to.tip, model, mm, my.init.fun.BM = NULL, my.lower.BM = NULL, my.upper.BM = NULL, my.init.fun.OU = NULL, my.lower.OU = NULL, my.upper.OU = NULL, n.bs = 100, resim = T, verbose = F){
  # simulate 
  if(resim){
    sim.z <- get.simulated.trait(VL = VL, pars = pars, vars = vars, model = model, mm = mm, n.sim = n.bs)
  } else {
    sim.z <- VL
  }
  if(n.bs == 1) {
    namesz <- names(sim.z)
    sim.z <- matrix(sim.z, nrow = 1)
    colnames(sim.z) <- namesz
  }
  bs.results <- c()
  for(ii in 1:n.bs){ # 10 minutes for 100 replicates
    if(verbose) cat("bootstrap ", model, " #", ii, "\n")
    if(model == "OU"){
      opt_sim <- optim.fun.repeated(3, lik.OU, my.init.fun.OU,
                                    tree = tree, VL = sim.z[ii,], D = D, root.to.tip = root.to.tip, mm = mm,# D has to be passed on, and root to tip as well
                                    lower = my.lower.OU, upper = my.upper.OU)
      par.sim <- lik.OU(par = opt_sim$pars, tree = tree, VL = sim.z[ii,], D = D, root.to.tip = root.to.tip, full.output = T, mm = mm)$beta # the ML ancestral trait Y_A and theta in this order
    }
    if(model == "BM"){
      opt_sim <- optim.fun.repeated(3, lik.BM, my.init.fun.BM, tree = tree, VL = sim.z[ii,], mm = mm,
                                    lower = my.lower.BM, upper = my.upper.BM)
      par.sim <- lik.BM(par = opt_sim$pars, tree = tree, VL = sim.z[ii,], full.output = T, mm = mm)$beta # the ML ancestral trait Y_A and theta in this order
    }
    
    h2.sim <- get.h2(pars = opt_sim$pars, vars= vars, model = model, mm = mm, fixed.effects = par.sim,  nsim = 1000, check.sim = F)
    bs.results <- rbind(bs.results, c(opt_sim$pars, opt_sim$lik, par.sim, h2.sim$anal.h2))
  }
  bs.results
}
# EXAMPLE USAGE
#do.bs(VL, opt1_2$pars, vars = list(vcv.mat), tree = treeBM, D = NULL, root.to.tip = NULL, model = "BM", mm = mm, n.bs = 100) -> truc
#do.bs(VL, opt2_2$pars, vars = list(vcv.mat, pat.dist.mat), tree = treeOU, D = D, root.to.tip = root.to.tip, model = "OU", mm = mm, n.bs = 100) -> truc2


optim.fun.repeated <- function(n.repeats, lik.fun, init.fun, verbose = F, lower = NULL, upper = NULL, ...){
  # a function to optimise properly the likelihood function
  # using repeated use of function optim
  generate.init.fun <- match.fun(init.fun)
  lik.fun <- match.fun(lik.fun)
  all.opt <- list()
  for(ii in 1:n.repeats){
    init <- generate.init.fun()
    
    if(length(init)==1){ # one parameter function, use optimise
      if(is.null(upper) | is.null(lower)){ # if unspecified bounds for parameters, use (0,1)
        interval <- c(0, 1)
      } else {
        interval <- c(lower[1], upper[1])
      }
      all.opt[[ii]] <- optimise(lik.fun, interval = interval, ...)
      names(all.opt[[ii]]) <- c("par", "value")
      all.opt[[ii]]$message <- "Successful convergence"
    } else { # otherwise use nmk
      if(is.null(upper) | is.null(lower)){ # if unspecified bounds for parameters, use nmk
        all.opt[[ii]] <- nmk(init, lik.fun, ...)
      } else {
        all.opt[[ii]] <- nmkb(init, lik.fun, lower = lower, upper = upper, ...)
      }
    }  
    
    if(verbose)print(ii)
  }
  if(verbose) cat(sum(unlist(lapply(all.opt, function(x)x$message == "Successful convergence"))), "optimisations converge\n")
  all.ll <- unlist(lapply(all.opt, function(x)x$value))
  min(all.ll) -> minll.pow
  if(verbose) cat(length(which(all.ll < minll.pow + 1 & all.ll > minll.pow)), " optim within 1 log lik of minimum likelihood\n")
  output <- list(minll.pow, all.opt[[which.min(all.ll)]]$par)
  names(output) <- c("lik", "pars")
  return(output)
}

init.white <- function()return(runif(1,0,1))
init.BM <- function(){
  return(runif(2, c(0.1, 0.1), c(0.3, 1)))
}
init.OU <- function(){
  return(runif(3, c(0, 0, 0), c(1, 1, 1))) # params in order VE, sigsq, alpha
}
init.BM3 <- function(){
  return(runif(4, c(0, 0, 0, 0), c(1, 1, 1, 0.4))) # params in order VE, sigsq0, sigsq1, tlim
}
init.BM.VEt <- function(){
  return(runif(3, c(0, 0, 0), c(1, 1, 0.01))) # params in order VE, sigsq, slope.VE
}
init.VEt <- function(){
  return(runif(2, c(0, 0.), c(1, 0.01)))  # params in order VE, slope.VE
}

# function for slight data cleaning
as.numeric.date <- function(date)1970 + as.numeric(as.Date(date))/365.25 # numeric date in years

var.pheno.sample <- function(idx, VL){ # returns phenotypic variance of pheno of sample of tips for the actual data
  return((mean(VL[idx]^2)-mean(VL[idx])^2))
}

# functions to assess goodness of fit

boot.pheno.sample <- function(idx, n.boot, VL){ # returns bootstrap intervals for phenotypic variance for the actual data
  var <- c()
  for(ii in 1:n.boot){
    vec <- sample(VL[idx], replace = T)
    var <- c(var, (mean(vec^2)-mean(vec)^2))
  }
  return(quantile(var,c(0.025, 0.975)))
}
exp.var.pheno.sample <- function(idx, sim.pheno){ # expected value of variance in a sample, computed from simulations
  return(mean(apply(sim.pheno[,idx], 1, function(vec)mean(vec^2)-mean(vec)^2)))
}
intervals.pred.var.pheno <- function(idx, sim.pheno){ # sd of variance in a sample, computed from simulations
  return(quantile(apply(sim.pheno[,idx], 1, function(vec)mean(vec^2)-mean(vec)^2), c(0.025,0.975)))
}
intervals.pred.h2 <- function(idx, sim.g, sim.e){ # sd of h2 in a sample, computed from simulations
  varg <- apply(sim.g[,idx], 1, var)
  vare <- apply(sim.e[,idx], 1, var)
  return(quantile(varg/(varg+vare), c(0.025,0.5,0.975)))
}
pred.var.pheno.sample <- function(idx, mu, Sigma){ # prediction for expected value of the variance for sample idx
  return(mean(mu[idx]^2) + mean(diag(Sigma)[idx]) - mean(outer(mu[idx], mu[idx], FUN = "*") + Sigma[idx,idx]))
}


##---------------------------------------------------------------------------------------------------------##
####                                         OTHER FUNCTIONS                                             ####
##---------------------------------------------------------------------------------------------------------##
is.phylo <- function(x) return("phylo" %in% class(x)) # is.phylo function from geiger package

get.VL.and.var <- function(VL.info, VL.measure, this.tree, extra.check = F){ # get VL data and tree descriptors
  
  VL <- VL.info[ , VL.measure]
  dates <- VL.info[ , "Date.sampled"]
  dates <- dates - min(dates, na.rm = T)
  names(VL) <- VL.info$ID
  
  # phylogenetic structure info:
  mrca.node <- find.mrca(this.tree)

  ntips <- length(this.tree$tip.label)
  root.to.mrca <- dist.nodes(this.tree)[mrca.node, ntips+1] # root to mrca distance (node ntips + 1 is root)
  vcv.mat <- vcv.phylo(this.tree) - root.to.mrca
  
  if(extra.check){ # check vcv.mat with alternative method
    vcv.mat2 <- find.vcv(this.tree, mrca.node)
    stopifnot(all(abs(range(vcv.mat2 - vcv.mat)) < 10^-10))
  }
  pat.dist.mat <- cophenetic(this.tree) # patristic distance between tips
  # return everything
  return(list(VL, dates, vcv.mat, pat.dist.mat))
}

get.h2 <- function(pars, vars, model, mm, fixed.effects, nsim = 1000, check.sim = F, return.full.sim = F){ # get h2 for a given VL and parameter, variables, and model
  n.obs  <- nrow(vars[[1]])
  if(!is.null(mm)) if(nrow(mm) != n.obs){print(nrow(mm)); print(n.obs); stop("model matrix does not have the right number of rows")}
  # get the Sigma and design matrix
  tmp <- create.SigmaG.SigmaE(pars, var = vars, model =  model, mm = mm)
  SigmaG <- tmp[[1]]
  SigmaE <- tmp[[2]]
  X <- tmp[[3]]
  mu <- X %*% fixed.effects # average at the tips according to least square regression
  
  # mu can be decomposed into the genetic component and the environmental (covariates) component
  if(model == "OU"){
    g <- X[,1:2] %*% fixed.effects[1:2,1]
    if(ncol(X) > 2){
      covariates <- X[, 3:ncol(X), drop = F] %*% fixed.effects[3:ncol(X), 1, drop = F]
    } else {
      covariates <- rep(0, n.obs)
    }
  }
  if(model == "BM"){
    g <- X[,1, drop = F] %*% fixed.effects[1,1]
    if(ncol(X) > 1){ # if covariates are present
      covariates <- X[, 2:ncol(X), drop = F] %*% fixed.effects[2:ncol(X), 1, drop = F]
    } else {
      covariates <- rep(0, n.obs)
    }
  }
  stopifnot(all(abs(g + covariates - mu) < 10^-10))
  
  # with covariates, where is the mu standing???
  
  if(check.sim){ # do simulations optionally
    sim.g <- mvrnorm(n = nsim, mu = g, Sigma = SigmaG,
                     tol = 1e-6, empirical = FALSE, EISPACK = FALSE)
    sim.e <- mvrnorm(n = nsim, mu = covariates, Sigma = SigmaE,
                     tol = 1e-6, empirical = FALSE, EISPACK = FALSE)
    sim.pheno <- sim.g + sim.e # 1000 trajectories * ntips
    
    VP <- apply(sim.pheno, 1, var)
    VG <- apply(sim.g, 1, var)
    VE <- apply(sim.e, 1, var)
    COVGE <- sapply(1:nsim, function(ii) cov(sim.g[ii, ], sim.e[ii,]))
    h2 <- VG/VP
    stopifnot(all(abs(VP - (VG + VE + 2*COVGE)) < 10^-10))
  }
  
  anal.mean.VG <- mean(diag(SigmaG)) - mean(SigmaG) + mean(g^2) - mean(outer(g, g))
  anal.mean.VE <- pars[1] * (1 - 1/n.obs) + mean(covariates^2) - mean(outer(covariates, covariates))
  # 
  # mean(VG) + mean(VE)
  
  if(return.full.sim & check.sim){
    return(list(sim.mean.VG = mean(VG), sim.mean.VE = mean(VE), sim.mean.COVGE = mean(COVGE), sim.h2 = mean(h2),
                anal.mean.VG = anal.mean.VG, anal.mean.VE = anal.mean.VE, anal.h2 = anal.mean.VG/(anal.mean.VG + anal.mean.VE), sim.g = sim.g, sim.e  = sim.e))
  }
  
  if(check.sim){
    return(list(sim.mean.VG = mean(VG), sim.mean.VE = mean(VE), sim.mean.COVGE = mean(COVGE), sim.h2 = mean(h2),
                anal.mean.VG = anal.mean.VG, anal.mean.VE = anal.mean.VE, anal.h2 = anal.mean.VG/(anal.mean.VG + anal.mean.VE)))
  } else {
    return(list(sim.mean.VG = NA, sim.mean.VE = NA, sim.mean.COVGE = NA, sim.h2 = NA,
                anal.mean.VG = anal.mean.VG, anal.mean.VE = anal.mean.VE, anal.h2 = anal.mean.VG/(anal.mean.VG + anal.mean.VE)))
  }
}
