rm(list = ls())
library(MASS)
library(plyr)
library(ape)

# Francois Blanquart 14.04.2017
# R code to reproduce the figures in "Viral genetic variation accounts for a third of variability in HIV-1 set-point viral load in Europe"

setwd("~/DID/BEEHIVE/GWAS/RGWAS/to_share/") # current directory
load("heritability_subtypeB_data.RData")

# mm                  model matrix of the phylogenetic linear model
# par.BM              inferred fixed effects of the BM phylogenetic linear model, for subtype B, GSVL (N = 1581)
# par.OU              inferred fixed effects of the OU phylogenetic linear model, for subtype B, GSVL (N = 1581)
# VL.info             1581 * 10 dataframe with individual information
#                       ID
#                       BEEHIVE_LVL gold standard viral load ('GSVL' in the paper)
#                       SPVL set-point viral load
#                       Date.sampled date of the sample
#                       t.mode transmission mode
#                       GENDER
#                       age.cat age category
#                       ethnic2 ethnicity
#                       assay.spvl2 assay used for most viral loads in SPVL
#                       assay.gsvl2 assay used for GSVL
# h2.BM               heritability in the BM model analysis for subtype B, GSVL (N = 1581)
# h2.OU               heritability in the OU model analysis for subtype B, GSVL (N = 1581)
# lm0                 NULL model, i.e. linear model without phylogenetic effect for subtype B, GSVL (N = 1581)
# opt1_2              output of the BM model analysis for subtype B, GSVL (N = 1581) ($pars are VE, sigmasq)
# opt2_2              output of the OU model analysis for subtype B, GSVL (N = 1581) ($pars are VE, sigmasq, alpha)
# subtree             phylogenetic tree of the 1581 sequences in 'ape' format

all(subtree$tip.label==VL.info$ID) # subtree and VL.info are ordered in same way

########################################       FIGURE 1B         #########################################
pat.dist.mat <- cophenetic.phylo(subtree)
vcv.mat <- vcv.phylo(subtree)

root.to.tip <- diag(vcv.mat)
VL.info$root.to.tip <- NA
VL.info$root.to.tip[match(rownames(vcv.mat), VL.info$ID)] <- root.to.tip

par(mar = c(4, 4, 2, 1) + 0.1) # bottom, left, top and right
# sort in bins:
nbins <- 8
bins <- quantile(root.to.tip, seq(0, 1, 1/(nbins)))
vcv.cat <- rep(NA, nrow(VL.info))
for(ii in 1:nbins){
  this.sub <- which(root.to.tip >= bins[ii] & root.to.tip < bins[ii + 1])
  vcv.cat[this.sub] <- ii
}

boxplot(VL.info$BEEHIVE_LVL ~ vcv.cat, at = (bins[1:nbins] + bins[2:(nbins+1)]) /2, boxwex = 0.005,
        xlim = c(0.15, 0.5), ylim = c(1.5, 7), axes = F, outline = F, lty = 1, col=rgb(0, 0, 0, alpha = 0.),
        ylab = "gold standard viral load (log10 copies/mL)", xlab = "root to tip distance", main = "Figure 1B")
axis(1, at = seq(0.15, 0.5, 0.05))
axis(2, at = seq(2, 7))

points(VL.info$root.to.tip, VL.info$BEEHIVE_LVL, pch = 20, col = 'gray', cex = 0.3)

boxplot(VL.info$BEEHIVE_LVL ~ vcv.cat, at = (bins[1:nbins] + bins[2:(nbins+1)]) /2, boxwex = 0.005,
        outline = F, lty = 1, add = T, axes = F, col=rgb(0, 0, 0, alpha = 0.))

########################################       FIGURE 1C         #######################################

get.corr <- function(list.of.pairs, this.VL){
  stopifnot(nrow(list.of.pairs)==2)
  cherry.VL <- apply(list.of.pairs, c(1,2), function(idx) this.VL[idx])
  #cor(cherry.VL[1,], cherry.VL[2,])
  n <- ncol(cherry.VL)
  r <- cor(cherry.VL[1,], cherry.VL[2,]) # correlation
  z <- 0.5 * log((1 + r) / (1 - r))
  ci.z <- c(z - 1.96 * sqrt(1/(n-3)), z,  z + 1.96 * sqrt(1/(n-3)))
  ci.r <- (exp(2 * ci.z) - 1) / (exp(2 * ci.z) + 1)
  # cor.test(cherry.VL[1,], cherry.VL[2,]) # the R way to do it yields same result
  #lm(cherry.VL[1,] ~  cherry.VL[2,])
  return(ci.r) # CORRELATION COEFFICIENT AND CONFIDENCE INTERVAL
}
get.stochastic.cov <- function(one.pair, mat.sim.VL){ # get stochastic covariance (across simulations) for one pair
  nsim <- nrow(mat.sim.VL)
  stopifnot(length(one.pair) == 2)
  sim.pair <- mat.sim.VL[, one.pair]
  stopifnot(ncol(sim.pair) == 2)
  return(cov(sim.pair[,1], sim.pair[,2]))
}
get.covariance <- function(list.of.idx, this.VL){
  if(!is.null(dim(list.of.idx))){
    if(nrow(list.of.idx) > 1){
      coeff <- cov(this.VL[list.of.idx[,1]],  this.VL[list.of.idx[,2]])
    } else {
      coeff <- NA
    }
  } else {
    coeff <- NA
  }
  return(coeff)
}
# identify cherries:
n <- length(subtree$tip.label)
cherry.nodes <- which(tabulate(subtree$edge[, 1][subtree$edge[, 2] <= n]) == 2)
cherry.tips <- sapply(cherry.nodes, function(node.nb) { # pairs of tips
  node.idx <- which(subtree$edge[,1] == node.nb)
  stopifnot(length(node.idx) == 2)
  return(subtree$edge[node.idx, 2])
})
print("number of cherries: ")
print(ncol(cherry.tips))
pd.cherry <- sapply(1:ncol(cherry.tips), function(ii) pat.dist.mat[cherry.tips[1,ii], cherry.tips[2,ii]]) # patristic distance of cherries
vcv.cherry <- sapply(1:ncol(cherry.tips), function(ii) vcv.mat[cherry.tips[1,ii], cherry.tips[2,ii]]) # var-cov distance of cherries

# covariance in simulation

nsim <- 10000

if(!exists("sim.VL.OU")) sim.VL.OU <- get.simulated.trait(VL.info$BEEHIVE_LVL, pars = opt2_2$pars, vars = list(vcv.mat, pat.dist.mat), model = "OU", mm = mm, n.sim = nsim)
if(!exists("sim.VL.BM")) sim.VL.BM <- get.simulated.trait(VL.info$BEEHIVE_LVL, pars = opt1_2$pars, vars = list(vcv.mat), model = "BM", mm = mm, n.sim = nsim)

# compute the correlation for different thresholds on patristic distance

tau_seq <- quantile(pd.cherry, seq(0, 1, 1/10))
if(!exists("all.cov") | !exists("sim.cov.OU") | !exists("sim.cov.BM")){
  all.cov <- c()
  sim.cov.OU <- c()
  sim.cov.BM <- c()
  for(ii in 1:(length(tau_seq) - 1)){
    this.pd <- which(pd.cherry >= tau_seq[ii] & pd.cherry < tau_seq[ii + 1])
    if(length(this.pd) > 10){
      print(length(this.pd))
      all.cov <- rbind(all.cov, get.corr(cherry.tips[, this.pd], VL.info$BEEHIVE_LVL))
      sim.cov.BM.all <- sapply(1:nsim, function(ii) get.corr(cherry.tips[, this.pd], sim.VL.BM[ii,])[2])
      sim.cov.OU.all <- sapply(1:nsim, function(ii) get.corr(cherry.tips[, this.pd], sim.VL.OU[ii,])[2])
      sim.cov.BM <- c(sim.cov.BM, median(sim.cov.BM.all))
      sim.cov.OU <- c(sim.cov.OU, median(sim.cov.OU.all))
    } else {
      all.cov <- rbind(all.cov, c(NA, NA, NA))
      sim.cov.BM <- c(sim.cov.BM, NA)
      sim.cov.OU <- c(sim.cov.OU, NA)
    }  
  }
}

# ... AND THE PLOT:

tau_mid <- 0.5 * (tau_seq[2:(length(tau_seq))] + tau_seq[1:(length(tau_seq) - 1)])
plot(tau_mid, all.cov[,2], type = "b", pch = 20, xlim = c(-0.01, 0.4),
     ylim = c(-0.5, 0.7), ylab = c("correlation coefficient"), xlab = "patristic distance between cherries", axes = F, main = "Figure 1C")
axis(1, at = seq(0, 0.4, 0.05))
axis(2, at = seq(-0.5, 0.7, 0.2))

for(ii in 1:nrow(all.cov)) segments(tau_mid[ii], all.cov[ii, 1], tau_mid[ii], all.cov[ii, 3])

f <- function(x, a, b) {a * exp(b * x)}
nlm0 <- nls(all.cov[,2] ~ f(tau_mid, a, b), start = c(a = 1, b = 1)) # exponential fit

points(tau_mid, sim.cov.BM, lwd = 3, col = "blue", type = "l")
points(tau_mid, sim.cov.OU, lwd = 3, col = "red", type = "l")

points(0, median(h2.OU$anal.h2), col = "red", cex = 2, pch = 20)
points(0, median(h2.BM$anal.h2), col = "blue", cex = 2, pch = 20)

points(x <- seq(0, 0.4, 0.01), f(x, coef(nlm0)['a'], coef(nlm0)['b']), lty = 2, type = "l") # exponential fit

abline(a = lm0$coefficients[1], b = lm0$coefficients[2], lty= 2)
abline(h = 0, lty = 1)
legend("topright", col = c("blue", "red"), lwd = 3, lty = 1, legend = c("BM", "OU"))

########################################       FIGURE 2         #######################################

plot.h2 <- function(idx, y.pos){
  h2 <- data$h2..low.high.[idx]
  h2 <- gsub("]", "", h2)
  h2 <- gsub("\\[", "", h2)
  h2 <- gsub("e-", "10minus", h2) # here tricky, needs to keep the e- so we temporarily replace with 10minus
  h2 <- gsub("; ", " ", h2)
  h2 <- gsub("10minus", "e-", h2)
  h2 <- as.numeric(unlist(strsplit(h2, " ")))
  
  points(h2[1], y.pos , pch = 20)
  segments(h2[2], y.pos, h2[3], y.pos)
}
data <- read.table("all.summary2017-03-10.csv", header = T, sep = ",")

View(data)
list.of.idx <- c(14, 15, 32, 33, 69, 72, 66, 54, 60, 48, 42) # lines that we will plot
ymax <- length(list.of.idx) + 1

plot(NULL, xlab = "", ylab = "", axes = F, xlim = c(-1,1), ylim = c(0, ymax), main = "Figure 2")

rect(-1, 4.6, 1.02, 7.4, col = 'gray90')
rect(-1, 0.6, 1.02, 4.4, col = 'gray70')

axis(1, seq(0, 1,0.2), labels = seq(0, 100, 20))
mtext("heritability (%)", side = 1, at = c(0.5), line = 2.5)

for(colname in c("model", "country", "measure", "subtype", "gene", "gender", "transmission.mode")) data[, colname] <- as.character(data[,colname])
for(ii in 1:length(list.of.idx)){
  idx <- list.of.idx[ii]
  label <- paste(data[idx, c("model", "country", "measure",  "gene", "gender", "transmission.mode")], collapse = " ")
  label <- gsub(" -", "", label)
  label <- gsub("LVL", "GSVL", label)
  label <- gsub(" ", ", ", label)
  label <- gsub(", all", "", label)
  label <- gsub(", full", "", label)
  label <- paste0(label, " (N = ", data[idx, 'N'], ")")
  text(x = -1, y = ymax - ii, label, pos = 4)
  plot.h2(idx, ymax - ii)
}
abline(v = 0, lty = 2)
abline(v = 1, lty = 2)

########################################       FIGURE 3         #######################################

load("heritability_subtypeB_along_genome.RData") # data on h2 along the genome
# list.of.ci              list of confidence intervals for heritability for BM (first two columns) and OU (last two columns) for 17 windows along the genome
# list.of.h2              list of heritabilities for BM (first column) and OU (second column) for 17 windows along the genome
# h2.ci.env               confidence intervals for heritability in the env gene under OU
# h2.ci.gag               confidence intervals for heritability in the gag gene under OU
# h2.ci.pol               confidence intervals for heritability in the pol gene under OU
# h2.env                  heritability in the env gene under OU
# h2.gag                  heritability in the gag gene under OU
# h2.pol                  heritability in the pol gene under OU
# list.of.tree            list of the 17 phylogenetic trees for the 17 windows across the genome in 'ape' format
# tree.env                phylogenetic tree of the env gene in 'ape' format
# tree.gag                phylogenetic tree of the gag gene in 'ape' format
# tree.pol                phylogenetic tree of the pol gene in 'ape' format

# index gag in the alignment: 397-2022
idx.g1 <- 397
idx.g2 <- 2022
# idx pol:
idx.p1 <- 1710
idx.p2 <- 4263
# idx env:
idx.e1 <- 5537
idx.e2 <- 8323

# length of alignment
n.tot <- 9181
index.pos <- seq(500, 8500, 500)

plot(NULL,  ylim = c(0,0.4), xlim = c(0, 9181), xlab = "position in alignment (bp)", ylab = "heritability", axes = F, xaxs="i", yaxs = "i", main = "Figure 3")
axis(1, c(1, 2500, 5000, 7500, 9181))
axis(2, seq(0, 0.4, 0.1))

polygon(c(index.pos, rev(index.pos)), c(list.of.ci[, 1], rev(list.of.ci[, 2])), col = rgb(0,0,0,0.1), border = NA)
polygon(c(index.pos, rev(index.pos)), c(list.of.ci[, 3], rev(list.of.ci[, 4])), col = rgb(0,0,0,0.1), border = NA)

polygon(c(idx.g1, idx.g2, idx.g2, idx.g1), c(h2.ci.gag[1], h2.ci.gag[1], h2.ci.gag[2], h2.ci.gag[2]), col = rgb(0,0,1,0.1), border = NA)
polygon(c(idx.p1, idx.p2, idx.p2, idx.p1), c(h2.ci.pol[1], h2.ci.pol[1], h2.ci.pol[2], h2.ci.pol[2]), col = rgb(0,1,0,0.1), border = NA)
polygon(c(idx.e1, idx.e2, idx.e2, idx.e1), c(h2.ci.env[1], h2.ci.env[1], h2.ci.env[2], h2.ci.env[2]), col = rgb(1,0,0,0.1), border = NA)

points(index.pos, list.of.h2[,1], type = "o", pch = 20, col = "gray56", lwd = 2)
points(index.pos, list.of.h2[,2], type = "o", pch = 20, lwd = 2)

abline(h = 0.31, lty = 2, lwd = 2) # heritability OU full genome
abline(h = 0.18, lty = 2, col = "gray56", lwd = 2) # heritability BM full genome

segments(idx.g1, h2.gag, idx.g2, h2.gag, col = "blue", lwd = 2)
segments(idx.p1, h2.pol, idx.p2, h2.pol, col = "green3", lwd = 2)
segments(idx.e1, h2.env, idx.e2, h2.env, col = "red", lwd = 2)


########################################       FIGURE S2         #######################################

all.results <- read.table('table_sampling_size.csv', header = T, sep = ",")

plot(NULL, ylim = c(0,1), xlim = c(0, 1600), xlab = "sample size", ylab = "estimated heritability BM", main = "Figure S2A")
to.plot.ci <- which(all.results$fraction >= 0.)
for(ii in 1:length(to.plot.ci)) segments(all.results$n[to.plot.ci[ii]], all.results$h2_BM_low[to.plot.ci[ii]], all.results$n[to.plot.ci[ii]], all.results$h2_BM_high[to.plot.ci[ii]], col = "gray")
segments(1581, 0.08, 1581, 0.26, col = "gray")
points(all.results$n, all.results$h2_BM, pch = 20)
points(1581, 0.17, pch = 20) # add best estimate
abline(h = 0.17, lty = 2)

plot(NULL, pch = 20, ylim = c(0,1), xlim = c(0, 1600), xlab = "sample size", ylab = "estimated heritability OU", main = "Figure S2B")
for(ii in 1:length(to.plot.ci)) segments(all.results$n[to.plot.ci[ii]], all.results$h2_OU_low[to.plot.ci[ii]], all.results$n[to.plot.ci[ii]], all.results$h2_OU_high[to.plot.ci[ii]], col = "gray")
segments(1581, 0.15, 1581, 0.43, col = "gray")
points(all.results$n, all.results$h2_OU, pch = 20)
points(1581, 0.31, pch = 20)
abline(h = 0.31, lty = 2)


########################################       FIGURE S3         #######################################

setwd("~/DID/BEEHIVE/GWAS/RGWAS/to_share/")

par.BM.GSVL <- c(0.4291847, 0.4984819)
par.OU.GSVL <- c(0.3479514, 3.0252193, 8.4619140)
par.BM.SPVL <- c(0.4167835, 0.3522552)
par.OU.SPVL <- c(0.3720972, 1.7478657, 7.6119359)

OU.GSVL.sim <- get.h2(pars = par.OU.GSVL, vars= list(vcv.mat, pat.dist.mat), model = "OU", mm = mm, fixed.effects = par.OU, nsim = 1000, check.sim = T, return.full.sim = T)
BM.GSVL.sim <- get.h2(pars = par.BM.GSVL, vars= list(vcv.mat), model = "BM", mm = mm, fixed.effects = par.BM, nsim = 1000, check.sim = T, return.full.sim = T)

OU.SPVL.sim <- get.h2(pars = par.OU.SPVL, vars= list(vcv.mat, pat.dist.mat), model = "OU", mm = mm, fixed.effects = par.OU, nsim = 1000, check.sim = T, return.full.sim = T)
BM.SPVL.sim <- get.h2(pars = par.BM.SPVL, vars= list(vcv.mat), model = "BM", mm = mm, fixed.effects = par.BM, nsim = 1000, check.sim = T, return.full.sim = T)

# check that row names in mat correspond to VL.info
stopifnot(row.names(vcv.mat) == VL.info$ID)

VL.info$Date.cat <- floor(VL.info$Date.sampled / 2) * 2 # date categories

BM.GSVL.sim$sim.p <- BM.GSVL.sim$sim.g + BM.GSVL.sim$sim.e
OU.GSVL.sim$sim.p <- OU.GSVL.sim$sim.g + OU.GSVL.sim$sim.e
BM.SPVL.sim$sim.p <- BM.SPVL.sim$sim.g + BM.SPVL.sim$sim.e
OU.SPVL.sim$sim.p <- OU.SPVL.sim$sim.g + OU.SPVL.sim$sim.e

# mean variances by date category

all.var <- c()
for(year.cat in sort(unique(VL.info$Date.cat))){
  n <- sum(VL.info$Date.cat == year.cat)
  mean.varg.BM.GSVL <- mean(apply(BM.GSVL.sim$sim.g[, which(VL.info$Date.cat == year.cat)], 1, var))
  mean.varg.OU.GSVL <- mean(apply(OU.GSVL.sim$sim.g[, which(VL.info$Date.cat == year.cat)], 1, var))
  
  mean.varg.BM.SPVL <- mean(apply(BM.SPVL.sim$sim.g[, which(VL.info$Date.cat == year.cat)], 1, var))
  mean.varg.OU.SPVL <- mean(apply(OU.SPVL.sim$sim.g[, which(VL.info$Date.cat == year.cat)], 1, var))
  
  mean.varp.BM.GSVL <- mean(apply(BM.GSVL.sim$sim.p[, which(VL.info$Date.cat == year.cat)], 1, var))
  mean.varp.OU.GSVL <- mean(apply(OU.GSVL.sim$sim.p[, which(VL.info$Date.cat == year.cat)], 1, var))
  
  mean.varp.BM.SPVL <- mean(apply(BM.SPVL.sim$sim.p[, which(VL.info$Date.cat == year.cat)], 1, var))
  mean.varp.OU.SPVL <- mean(apply(OU.SPVL.sim$sim.p[, which(VL.info$Date.cat == year.cat)], 1, var))
  
  var.GSVL <- var(VL.info$BEEHIVE_LVL[which(VL.info$Date.cat == year.cat)])
  var.SPVL <- var(VL.info$SPVL[which(VL.info$Date.cat == year.cat)])
  
  all.var <- rbind(all.var, c(year.cat, n, mean.varg.BM.GSVL, mean.varg.OU.GSVL, mean.varg.BM.SPVL, mean.varg.OU.SPVL,
                              mean.varp.BM.GSVL, mean.varp.OU.GSVL, mean.varp.BM.SPVL, mean.varp.OU.SPVL, var.GSVL, var.SPVL))
}
all.var <- data.frame(all.var)
names(all.var) <- c('year.cat', 'n', 'mean.varg.BM.GSVL', 'mean.varg.OU.GSVL', 'mean.varg.BM.SPVL', 'mean.varg.OU.SPVL', 'mean.varp.BM.GSVL', 'mean.varp.OU.GSVL', 'mean.varp.BM.SPVL', 'mean.varp.OU.SPVL', 'var.GSVL', 'var.SPVL')

plot(all.var$year.cat, all.var$var.GSVL, pch = 20, ylim = c(0, 1), type = "b", xlab = "year", ylab = "genetic and phenotypic variance", main = "Figure S3")
points(all.var$year.cat, all.var$var.SPVL, pch = 20, ylim = c(0, 1), type = "b", col = "gray")
points(all.var$year.cat, all.var$mean.varg.BM.GSVL, type = "l", lty = 2, lwd = 2)
points(all.var$year.cat, all.var$mean.varg.OU.GSVL, type = "l", col = "black", lty = 3, lwd = 2)
points(all.var$year.cat, all.var$mean.varg.BM.SPVL, type = "l", col = "gray", lty = 2, lwd = 2)
points(all.var$year.cat, all.var$mean.varg.OU.SPVL, type = "l", col = "gray", lty = 3, lwd = 2)

lm.GSVL <- lm(all.var$var.GSVL ~ all.var$year.cat, weights = all.var$n)
lm.SPVL <- lm(all.var$var.SPVL ~ all.var$year.cat, weights = all.var$n)

abline(lm.GSVL$coefficients[1], lm.GSVL$coefficients[2], col = 'black', lty = 1, lwd = 1)
abline(lm.SPVL$coefficients[1], lm.SPVL$coefficients[2], col = 'gray', lty = 1, lwd = 1)

# legend
legend("topleft", legend = c("data GSVL", "data SPVL", "linear regression data GSVL", "linear regression data SPVL", "genetic variance BM GSVL", "genetic variance OU GSVL", "genetic variance BM SPVL", "genetic variance OU SPVL"),
       col = c("black", "gray", "black", "gray", "black", "black", "gray", "gray"), lty  = c(1, 1, 1, 1, 2, 3, 2, 3), lwd = c(1, 1, 1, 1, 2, 2, 2, 2), pch = c(20, 20, NA, NA, NA, NA, NA, NA))

# rate of variance evolution
(all.var$mean.varg.BM.GSVL[nrow(all.var)] - all.var$mean.varg.BM.GSVL[1]) / (all.var$year.cat[nrow(all.var)] - all.var$year.cat[1])
(all.var$mean.varg.OU.GSVL[nrow(all.var)] - all.var$mean.varg.OU.GSVL[1]) / (all.var$year.cat[nrow(all.var)] - all.var$year.cat[1])

########################################       FIGURE S4         #######################################

df2 <- read.table("linkage_disequilibrium.csv", header = T, sep = ",") # table of linkage disequilibrium for 100000 pairs of randomly chosen positions
sum.df2 <- ddply(df2, .(dist.cat), summarise, mean.LD = mean(LD, na.rm = T))

plot(NULL, type = "b", ylim = c(-0.02, 0.1), xlim = c(0, 3), ylab = "linkage disequilibrium", xlab = "distance (bp)", axes = F, main = "Figure S4")
axis(side = 1, at = seq(0, 3, 1), labels = c(1, 10, 100, 1000))
axis(side = 2, at = seq(-0.02, 0.1, 0.02))
points(log10(sum.df2$dist.cat), sum.df2$mean.LD, pch = 20, type = "o")
abline(h = 0, lty = 2)




