# Francois Blanquart adaptation of Lam Ho's code
transf.branch.lengths <- function (phy, model = c("BM", "OUrandomRoot", "OUfixedRoot", 
    "lambda", "kappa", "delta", "EB", "trend"), parameters = NULL, 
    check.pruningwise = TRUE, check.ultrametric = TRUE, D = NULL, 
    check.names = TRUE) 
{
    if (!inherits(phy, "phylo")) 
        stop("object \"phy\" is not of class \"phylo\".")
    model = match.arg(model)
    if (model == "trend") 
        if (is.ultrametric(phy)) 
            stop("the trend is unidentifiable for ultrametric trees.")
    if (is.null(phy$edge.length)) 
        stop("the tree has no branch lengths.")
    if (is.null(phy$tip.label)) 
        stop("the tree has no tip labels.")
    tol = 1e-10
    if (check.pruningwise) 
        phy = reorder(phy, "pruningwise")
    n <- length(phy$tip.label)
    N <- dim(phy$edge)[1]
    ROOT <- n + 1L
    anc <- phy$edge[, 1]
    des <- phy$edge[, 2]
    externalEdge = (des <= n)
    if (!is.null(phy$root.edge)) 
        if (phy$root.edge > 0) 
            stop("the tree is supposed to have no root edge (or of length 0).")
    parameters.default = c(1/2, 1, 1, 1, 0, 0, 1) # FB added 1 for sigma2; NOTE ALPHA PARAMETER HAS TO BE 1/2 BY DEFAULT for BM
    names(parameters.default) = c("alpha", "lambda", "kappa", 
        "delta", "rate", "sigma2_error", "sigma2") # FB added sigma2
    if (is.null(parameters)) {
        parameters = parameters.default
    }
    else {
        if (class(parameters) != "list") {
            stop("please specify parameters as a list().")
        }
        else {
            specified <- !c(is.null(parameters$alpha), is.null(parameters$lambda), 
                is.null(parameters$kappa), is.null(parameters$delta), 
                is.null(parameters$rate), is.null(parameters$sigma2_error), is.null(parameters$sigma2)) # FB added  sigma2
            parameters.user <- c(parameters$alpha, parameters$lambda, 
                parameters$kappa, parameters$delta, parameters$rate, 
                parameters$sigma2_error, parameters$sigma2)
            names(parameters.default) = c("alpha", "lambda", 
                "kappa", "delta", "rate", "sigma2_error", "sigma2") # FB added sigma2
            parameters <- parameters.default
            parameters[specified] <- parameters.user
        }
    }
    p = list(alpha = parameters[1], lambda = parameters[2], kappa = parameters[3], 
        delta = parameters[4], rate = parameters[5], sigma2_error = parameters[6], sigma2 = parameters[7]) # FB added sigma2

    root.edge = 0
    diagWeight = rep(1, n)
    OU = c("OUrandomRoot", "OUfixedRoot")
    errEdge = rep(p$sigma2_error * 2 * p$alpha / p$sigma2, n) # FB added * 2 * p$alpha / p$sigma2 compared to Lam Ho version; for BM model this is 1/p$sigma2 as it should, because default alpha is 1/2
    
    if (model %in% c("BM", "trend")) {
        edge.length = phy$edge.length
        stopifnot(p$alpha == 1/2) # for BM model p$alpha has to be 1.2
    }    
    if (model %in% OU) {
        if (check.ultrametric) {
            D = numeric(n)
            if (!is.ultrametric(phy)) {
                flag = 1
                dis = pruningwise.distFromRoot(phy)
                D = max(dis[1:n]) - dis[1:n]
                D = D - mean(D)
                phy$edge.length[externalEdge] <- phy$edge.length[externalEdge] + 
                  D[des[externalEdge]]
            }
        }
        else {
            if (is.null(D)) 
                stop("Provide D if you choose check.ultrametric=F")
            if (length(D) != n) 
                stop("D should be a vector with one term for each tip in the tree")
            if (check.names) {
                if (is.null(names(D))) 
                  stop("D is lacking names (tip labels)")
                ordr = match(phy$tip.label, names(D))
                if (sum(is.na(ordr)) > 0) 
                  stop("names of D do not match the tree tip labels.")
                D = D[ordr, drop = F]
            }
        }
        times <- pruningwise.branching.times(phy)
        Tmax <- max(times)
        alpha = p$alpha
        errEdge = errEdge * exp(- 2 * alpha * D[des[externalEdge]])  # FB changed to - 2 * alpha * ...
        if (model == "OUrandomRoot") {
            distFromRoot <- exp(-2 * alpha * times)
            d1 = distFromRoot[anc - n]
            d2 = numeric(N)
            d2[externalEdge] = exp(-2 * alpha * D[des[externalEdge]])
            d2[!externalEdge] = distFromRoot[des[!externalEdge] - 
                n]
        }
        if (model == "OUfixedRoot") {
            distFromRoot <- exp(-2 * alpha * times) * (1 - exp(-2 * 
                alpha * (Tmax - times)))
            d1 = distFromRoot[anc - n]
            d2 = numeric(N)
            d2[externalEdge] = exp(-2 * alpha * D[des[externalEdge]]) * 
                (1 - exp(-2 * alpha * (Tmax - D[des[externalEdge]])))
            d2[!externalEdge] = distFromRoot[des[!externalEdge] - 
                n]
        }
        edge.length = d2 - d1
        root.edge = min(distFromRoot)
        diagWeight = exp(alpha * D)
    }
    if (model == "lambda") {
        lambda = p$lambda
        distFromRoot <- pruningwise.distFromRoot(phy)
        edge.length = phy$edge.length * lambda
        edge.length[externalEdge] = edge.length[externalEdge] + 
            (1 - lambda) * distFromRoot[des[externalEdge]]
    }
    if (model == "kappa") {
        kappa = p$kappa
        edge.length = phy$edge.length^kappa
    }
    if (model == "delta") {
        delta = p$delta
        distFromRoot <- pruningwise.distFromRoot(phy)
        depth = max(distFromRoot)
        edge.length = (distFromRoot[des]^delta - distFromRoot[anc]^delta) * 
            depth^(1 - delta)
    }
    if (model == "EB") {
        rate = p$rate
        if (rate == 0) 
            edge.length = phy$edge.length
        else {
            distFromRoot <- pruningwise.distFromRoot(phy)
            edge.length = (exp(rate * distFromRoot[des]) - exp(rate * 
                distFromRoot[anc]))/rate
        }
    }
    edge.length[externalEdge] = edge.length[externalEdge] + errEdge
    phy$edge.length = edge.length
    phy$root.edge = root.edge
    names(diagWeight) = phy$tip.label
    return(list(tree = phy, diagWeight = diagWeight))
}
